local M = {}

local factors            = require("scenario/gravitationalRacing/dataValues/factors")
local tableComp          = require("scenario/gravitationalRacing/utils/tableComprehension")
local ClassVector        = require("scenario/gravitationalRacing/classes/classVector")
local ClassVehicle       = require("scenario/gravitationalRacing/classes/classVehicle")
local ClassCelestial     = require("scenario/gravitationalRacing/classes/classCelestial")
local textureHandler     = require("celestialCreator/textureHandler")
local stringFormatter    = require("scenario/gravitationalRacing/utils/stringFormatting")
local celestialsHandler  = require("scenario/gravitationalRacing/celestial/celestialsHandler")
local checkpointsHandler = require("scenario/gravitationalRacing/scenario/track/checkpointsHandler")

local pack = table.pack or function(...) return {n=select('#',...),...} end

local initialised = false
local showAxis = false
local vehicles = {}
local previewData = {shouldUpdate = false, pos = nil, scale = 0, celestialType = nil, linearPathData = nil, binary = false}

--Stores the mapping from UI celestial name (ex: Jupiter (1)) -> obj name in-world (ex: planet_24)
local nameObjMappings = {}
--Stores the number of each celestial in world
local nameNumbers = {}

--Stores the orbit ranges that are taken around a particular celestial
local orbitGaps = {}

local function formatComponentSystem(c1Name, c2Name)
  --[[
  Combines two components of a binary system to a formatted system name
  ]]--
  return c1Name.." | "..c2Name
end

local function unformatObjName(name)
  --[[
  Converts a formatted name into its unformatted counter-part
  For example:
  -Neutron Star (1) -> star_3
  -Black Hole 1|2 -> blackhole_1, blackhole_2
  ]]--
  if not name then
    log("E", "gr:unformatObjName()", "name is nil")
    return
  end

  --Indicates it is a binary system
  if name:find("|") then
    local componentNames = {}
    --Matches: <celestialModel> [<id>]
    for token in name:gmatch("((%w+)( )(%[%d+%]))", "") do
      table.insert(componentNames, nameObjMappings[token])
    end

    return componentNames[1], componentNames[2]
  end

  local objName = nameObjMappings[name]
  if not objName then
    error("unformatObjName(): name="..name.." does not exist in mappings")
  end

  return objName
end

local function formatObjName(obj)
  --[[
  Returns the formatted name from the obj name
  ]]--
  if not obj then
    error("gr:formatObjName(): obj is nil")
  end

  for formatName, objName in pairs(nameObjMappings) do
    if obj == objName then
      return formatName
    end
  end

  log("E", "gr:formatObjName()", "Cannot find formatted name with objName="..obj)
end

local function updateAPreview(pos, scale, celestialType, linearPathData, id)
  --[[
  Updates a specific preview
  ]]--
  local name = "preview"..id
  --Black holes and Saturn both have custom hologram models
  local type = (celestialType == "blackhole" or celestialType == "saturn") and celestialType or "sphere"

  local obj = scenetree.findObject(name)

  --The chosen celestial has a different model to the preview, so remove it to be recreated correctly later
  if obj and not obj.shapeName:find(type) then
    TorqueScript.eval(name..".delete();")
    obj = nil
  end

  if not obj then
    TorqueScript.eval([[
    new TSStatic(]]..name..[[) {
      shapeName = "levels/smallgrid/art/gravitationalRacing/holograms/]]..("hologram_"..type)..[[.dae";
      dynamic = "1";
      position = "]]..pos:getX()..[[ ]]..pos:getY()..[[ ]]..pos:getZ()..[[";
      scale = "]]..scale..[[ ]]..scale..[[ ]]..scale..[[";
    };
    ]])
  else
    TorqueScript.eval(name..'.position = "'..pos:getX()..' '..pos:getY()..' '..pos:getZ()..'";')
    TorqueScript.eval(name..'.scale = "'..scale..' '..scale..' '..scale..'";')
  end

  --Setup the preview for the path
  if linearPathData then
    if not linearPathData.axis or not linearPathData.length then
      return
    end

    local scales = {x = 2, y = 2, z = 2}
    scales[linearPathData.axis] = linearPathData.length * 2

    if not scenetree.findObject(name.."_path") then
      TorqueScript.eval([[
      new TSStatic(]]..name..[[_path) {
        shapeName = "levels/smallgrid/art/gravitationalRacing/trail.dae";
        dynamic = "1";
        position = "]]..pos:getX()..[[ ]]..pos:getY()..[[ ]]..pos:getZ()..[[";
        scale = "]]..scales.x..[[ ]]..scales.y..[[ ]]..scales.z..[[";
      };
      ]])
    else
      TorqueScript.eval(name..'_path.position = "'..pos:getX()..' '..pos:getY()..' '..pos:getZ()..'";')
      TorqueScript.eval(name..'_path.scale = "'..scales.x..' '..scales.y..' '..scales.z..'";')
    end
  end
end

local function getCelestialRange(orbit, celestialRadius)
  --[[
  Returns the upper and lower bound of a range of values that this celestial occupies
  in its orbit
  ]]--
  --Convert to integer form and always round up
  orbit           = math.floor(orbit           + 1)
  celestialRadius = math.floor(celestialRadius + 1)
  return {min = orbit - celestialRadius, max = orbit + celestialRadius}
end

local function getActualSaturnSize(scale)
  --[[
  Returns the actual radius of saturn, accounting for rings
  ]]--
  return scale * ClassCelestial.getRingRadius()
end

local function updatePreview(pos, scale, celestialType, linearPathData, binary)
  --[[
  Updates the preview's position, scale and render model
  If binary is true, the pos argument requires a table of two positions
  ]]
  if binary then
    updateAPreview(pos[1], scale, celestialType, nil, 1)
    updateAPreview(pos[2], scale, celestialType, nil, 2)
  else
    updateAPreview(pos, scale, celestialType, linearPathData, 1)

    --Delete the other preview in-case previous preview was binary
    if scenetree.findObject("preview2") then
      TorqueScript.eval("preview2.delete();")
    end
  end
end

local function resetPreview(isBinary, linearPathData)
  --[[
  Resets the preview back to its hide state
  ]]--
  local pos = ClassVector.new(10000, 0, 0)
  --Removes the preview from view
  if isBinary then
    updatePreview({pos, pos:clone()}, 0, "", nil, true)
  else
    updatePreview(pos, 0, "", linearPathData, false)
  end
end

local function getLinearPathData(pathData)
  --[[
  Returns a table with only path data relevant to linear path type
  ]]--
  if not pathData or not pathData.type == "linear" then
    return nil
  else
    return {axis = pathData.axis, length = pathData.length}
  end
end

local function getNextName(shapeName)
  --[[
  Returns the next name for this celestial type, and the id added to it
  ]]--
  local curr = nameNumbers[shapeName]
  local new = curr + 1

  nameNumbers[shapeName] = new
  return shapeName.." ["..new.."]", new
end

local function parsePathData(pathData)
  --[[
  Parses and formats path data to have default values where not specified
  ]]--
  --Not an error since there is a Possibility of no path data
  if not pathData then
    return
  end

  local formatPathData = {
    path = pathData.type,
    orbitRadius  = pathData.orbitRadius or 0,
    orbitingBody = pathData.orbitingBody,
    frequency    = pathData.frequency,
    length       = pathData.length,
    direction    = pathData.axis
  }

  return formatPathData
end

local function adjustOrbitRangesForParent(parent, range, add)
  --[[
  Adds or removes ranges which represent occupied orbital space around a celestial
  ]]--
  local orbits = orbitGaps[parent:getName()]
  if not orbits then
    if not add then
      --Pointless to remove ranges from a celestial that is not a parent
      return
    end

    orbitGaps[parent:getName()] = {
      --The surface of the parent will always be a lower bound for an orbit
      {min = 0, max = parent:getSurfaceOfCelestial()},
      {min = range.min, max = range.max}
    }
  else
    print("Orbits table:")
    for _, r in ipairs(orbits) do
      print(" {min="..r.min..", max="..r.max.."}")
    end
    print()

    if add then
      local index = nil
      for i, r in ipairs(orbits) do
        --If the current range is infront of the new one, slot the new in infront
        if r.min > range.max then
          index = i
          break
        end
      end

      if not index then
        --Indicative that this orbit is the outer most orbit, so will be appended to the end of the list
        index = #orbits + 1
      end

      --Add to the table
      table.insert(orbits, index, {min = range.min, max = range.max})
    else
      local index = nil
      --Find the index the value is at
      for i, r in ipairs(orbits) do
        --Could check max values as well but unecessary since two orbits cannot have the same min value
        --There is a 1m tolerance as when deleting a celestial, orbit will not be perfect - due to approximations in sin(x)
        if math.abs(r.min - range.min) < 1.01  then
          index = i
          break
        end
      end

      if not index then
        log("E", "adjustOrbitRangesForParent()", "Range={min="..range.min..", max="..range.max.."} does not exist in orbits tables!")
        return
      end

      --Remove from the table
      table.remove(orbits, index)
    end

    print("Orbits table (Ammended):")
    for _, r in ipairs(orbits) do
      print(" {min="..r.min..", max="..r.max.."}")
    end
    print()
  end
end

local function findLowestOrbit(parentCelestial, celestialRadius)
  --[[
  Returns the smallest distance this object can orbit at
  ]]
  --The radius will be 2m away from the objects touching
  local celestialSize = celestialRadius + 2

  local orbits = orbitGaps[parentCelestial:getName()]
  if not orbits then
    return parentCelestial:getSurfaceOfCelestial() + celestialSize
  end

  for i = 1, #orbits-1 do
    local range1, range2 = orbits[i], orbits[i+1]
    --Gap is the distance between the end of one orbit and the start of another
    local gap = range2.min - range1.max
    if gap > 2*(celestialRadius + 1) then
      --Position the celestial as close to the start of the gap as possible
      return range1.max + celestialSize
    end
  end

  --No gaps are large enough for this celestial, so return the largest orbit away
  --from the celestial + the distance for this celestial
  return orbits[#orbits].max + celestialSize
end

local function addOrbit(orbit, parentCelestial, celestialRadius)
  --[[
  Gets the next orbit for a celestial orbitting a parent
  ]]--
  local range = getCelestialRange(orbit, celestialRadius)
  --Add to the table
  adjustOrbitRangesForParent(parentCelestial, range, true)
end

local function getOrbitalPosition(orbitingBody, celestial, orbitRadius, radius, returnWorkings)
  --[[
  Returns the initial position the new celestial will start at around a parent
  returnWorkings is an optional parameter to return the orbit radius and the parent celestial
  ]]--
  local orbitingBodies = pack(unformatObjName(orbitingBody))
  local orbitingSystem = #orbitingBodies > 1

  orbitingBody    = orbitingSystem and celestialsHandler.convertToSystemName(orbitingBodies[1], orbitingBodies[2]) or orbitingBodies[1]
  parentCelestial = orbitingSystem and celestialsHandler.findSystem(orbitingBody) or celestialsHandler.findCelestial(orbitingBody)

  if not parentCelestial then
    return {"Failure", "Could not find parent celestial with name "..(orbitingBody or "nil")}
  end

  --Set center of rotation to the center celestial's coords
  initPos = parentCelestial:getPosition():clone()

  local celestialSize = celestial == "saturn" and getActualSaturnSize(scale) or radius
  local lowestOrbit = findLowestOrbit(parentCelestial, celestialSize)

  local finalOrbitRadius = orbitRadius > lowestOrbit and orbitRadius or lowestOrbit
  initPos:setY(initPos:getY() + finalOrbitRadius)

  if returnWorkings then
    return initPos, finalOrbitRadius, parentCelestial, celestialSize
  else
    return initPos
  end
end

local function showPreview(show, data)
  --[[
  Shows an outline of the celestial to be created
  If the celestial is orbiting a moving parent, sets shouldUpdate to true
  ]]--
  if not data then
    error("Something has gone badly wrong!")
  end

  local celestial = data.celestial
  local scale = data.radius/0.03

  local posData = data.position
  local pos = posData.coords
  local initPos = ClassVector.new(pos.x, pos.y, pos.z)
  local isBinary = data.class == "binary"

  local pathData = parsePathData(data.path)

  if not show then
    resetPreview(isBinary, getLinearPathData(pathData))
    previewData.shouldUpdate = false

    return {"Success"}
  end

  if pathData and pathData.path == "circular" then
    local orbitPos, orbitRadius, parentCelestial, _ = getOrbitalPosition(pathData.orbitingBody, celestial, pathData.orbitRadius, data.radius, true)
    initPos = orbitPos

    --TODO: When adding in super-systems, check if binary system center moves or not
    local movingParent
    if parentCelestial:instanceOf() == "ClassBinarySystem" then
      movingParent = false
    else
      movingParent = parentCelestial:getObjectType() == "fixedDynamic"
    end

    if movingParent then
      previewData = {
        shouldUpdate = true,
        posData = {centerCelestial = parentCelestial, orbitRadius = orbitRadius},
        scale = scale,
        celestialType = celestial,
        linearPathData = nil,
        binary = false
      }
    else
      previewData.shouldUpdate = false
    end
  else
    previewData.shouldUpdate = false

    if posData.type == "relative" or isBinary then
      local vehPos = be:getPlayerVehicle(0):getPosition()
      initPos = ClassVector.new(vehPos.x + pos.x, vehPos.y + pos.y, vehPos.z + pos.z)
    end
  end

  if isBinary then
    local minRadius = data.celestial == "blackhole" and scale * ClassCelestial.getAccretionDiscRadius() or data.radius
    if not orbitRadius or orbitRadius < minRadius then
      orbitRadius = minRadius
    end

    --Place each preview at opposite ends
    local position1 = ClassVector.new(initPos:getX(), initPos:getY() + orbitRadius, initPos:getZ())
    local position2 = ClassVector.new(initPos:getX(), initPos:getY() - orbitRadius, initPos:getZ())

    updatePreview({position1, position2}, scale, celestial, nil, true)
  else
    updatePreview(initPos, scale, celestial, linearPathData, false)
  end

  return {"Success"}
end

local function requestCreation(data)
  --[[
  Takes the data from the UI and creates one or more celestials based on that information
  NOTE: The data is gotten from the form that has checked that accceptable
  values have been entered, so is not necessary to do so here again
  ]]--
  if not data then
    error("Something has gone badly wrong!")
  end

  local celestial = data.celestial
  local type = ClassCelestial.getCelestialTypeFromModel(celestial)
  local shapeName = celestial == "blackhole"
                    and "blackhole_eventHorizon"
                    or  celestial

  local scale = data.radius / 0.03
  local objectType = data.objectType
  local passive = data.passive
  local delayed = false
  local rotation = nil
  local posData = data.position
  local pathData = parsePathData(data.path)

  local orbitRadius = pathData and pathData.orbitRadius
  local orbitingBody = pathData and pathData.orbitingBody

  local initPos = nil

  local isBinary = data.class == "binary"
  local isCircular = pathData and pathData.path == "circular"

  if isCircular then
    initPos, orbitRadius, parentCelestial, celestialRadius = getOrbitalPosition(orbitingBody, celestial, pathData.orbitRadius, data.radius, true)
    addOrbit(orbitRadius, parentCelestial, celestialRadius)
  else
    local coords = posData.coords
    if posData.type == "relative" or isBinary then
      local vehPos = be:getPlayerVehicle(0):getPosition()
      initPos = ClassVector.new(vehPos.x + coords.x, vehPos.y + coords.y, vehPos.z + coords.z)
    else
      initPos = ClassVector.new(coords.x, coords.y, coords.z)
    end
  end

  local celestialAtSamePlace = celestialsHandler.findCelestialMatch("getPosition", initPos, false)

  if #celestialAtSamePlace > 0 then
    return {"Failure", formatObjName(celestialAtSamePlace[1]).." already exists here!"}
  end

  if isBinary then
    local minRadius = type == "blackhole" and scale * ClassCelestial.getAccretionDiscRadius() or data.radius
    if not orbitRadius or orbitRadius < minRadius then
      orbitRadius = minRadius
    end

    local objs = {}

    for i = 1, 2 do
      local formattedName, id = getNextName(celestial)
      local objName = type.."_"..id

      -- Place each celestial at opposite ends
      local position = i == 1
                   and ClassVector.new(initPos:getX(), initPos:getY() + orbitRadius, initPos:getZ())
                    or ClassVector.new(initPos:getX(), initPos:getY() - orbitRadius, initPos:getZ())

      nameObjMappings[formattedName] = objName
      nameNumbers[celestial] = nameNumbers[celestial] + 1

      objs[i] = {formattedName = formattedName, position = position, objName = objName}
    end

    objs[1].instance = celestialsHandler.createCelestial(data.class, shapeName, objs[1].position, objs[1].objName, type, scale, objectType, objs[2].objName, passive, delayed, rotation, nil)
    objs[2].instance = celestialsHandler.createCelestial(data.class, shapeName, objs[2].position, objs[2].objName, type, scale, objectType, objs[1].objName, passive, delayed, rotation, nil)

    for i = 1, 2 do
      local currentObj, parentObj = objs[i].instance, objs[3-i].instance
      --Store a reference to the other for each component
      currentObj:addChild(parentObj)
      --Setup the velocity
      currentObj:setVelocity(celestialsHandler.getOrbitalVelocity(currentObj, parentObj))
    end

    local objBinaryName = celestialsHandler.convertToSystemName(objs[1].objName, objs[2].objName)
    celestialsHandler.createSystem(objBinaryName, {objs[1].instance, objs[2].instance}, pathData.frequency)

    --Reset preview
    resetPreview(true, nil)

    local formattedName1, formattedName2 = objs[1].formattedName, objs[2].formattedName
    --Return the conjoined name and the individual data
    return {"Success", type, {formatComponentSystem(formattedName1, formattedName2), formattedName1, formattedName2}}
  end

  --[[formatName is the UI's name (such as CHXR73 (2)),
  objName is the name it will be refered to internally (such as star_5)]]--
  local formatName, id = getNextName(celestial)
  local objName = celestialsHandler.findNextName(type)

  nameObjMappings[formatName] = objName

  local formattedParent = orbitingBody and nameObjMappings[orbitingBody]

  local object = celestialsHandler.createCelestial(data.class, shapeName, initPos, objName, type, scale, objectType, formattedParent, passive, delayed, rotation, pathData)

  if objectType == "static" then
    --When being created, the collisions are not set correctly automatically, so reload them
    be:reloadStaticCollision();
  else
    local parentCelestial = celestialsHandler.findCelestial(formattedParent)
    if isCircular then
      object:setVelocity(celestialsHandler.getOrbitalVelocity(object, parentCelestial))
      --Add a reference for this object to the parent
      parentCelestial:addChild(object)
    end

    object:setupPath(parentCelestial)
  end

  resetPreview(false, getLinearPathData(pathData))

  --Return the new celestial information
  return {"Success", type, {formatName}}
end

local function getCelestialsAssociated(parentInstance)
  --[[
  Returns all celestials that are associated with a parent
  This is not only the celestials orbiting this but the celestials orbiting children as well
  ]]--
  local childrenInstances = celestialsHandler.findCelestialMatch("getParentCelestial", parentInstance:getName(), true)
  local modifiedChildInst = {}

  for _, child in ipairs(childrenInstances) do
    --Do not allow binary systems to have their counter-part as an associate or this will create an infinite loop
    if parentInstance:instanceOf() == "ClassBinarySystem" or not (child:getParentCelestial() == parentInstance:getName() and parentInstance:getParentCelestial() == child:getName()) then
      table.insert(modifiedChildInst, child)
    end
  end

  local associates = {}

  for _, instance in ipairs(modifiedChildInst) do
    local instAssociates = getCelestialsAssociated(instance)
    --Add in each sub-associate and the child itself
    associates = tableComp.merge(associates, instAssociates)
    table.insert(associates, {instance = instance, isParent = #instAssociates > 0})
  end

  return associates
end

local function requestDeletion(name, binary, confirmed)
  --[[
  Deletes a celestial
  ]]--
  local children = {}
  local instance = nil

  if binary then
    local name1, name2 = unformatObjName(name)
    local inst1, inst2 = celestialsHandler.findCelestial(name1), celestialsHandler.findCelestial(name2)
    instance = celestialsHandler.findSystem(celestialsHandler.convertToSystemName(name1, name2))

    if not inst1 or not inst2 then
      return {"Failure", "Could not find celestial(s) with name(s) "..(name1 or "nil")..", "..(name2 or "nil")}
    end

    children = tableComp.mergeAppend(getCelestialsAssociated(inst1), getCelestialsAssociated(inst2), getCelestialsAssociated(instance))
  else
    local objName = unformatObjName(name)
    instance = celestialsHandler.findCelestial(objName)

    children = getCelestialsAssociated(instance)
  end

  if #children > 0 then
    --The user has confirmed to delete all children as well
    if confirmed then
      local namesRemoved, parentsRemoved = {}, {}

      if binary then
        local c1, c2 = unformatObjName(name)
        c1, c2 = formatObjName(c1), formatObjName(c2)

        namesRemoved = {name, c1, c2}
        parentsRemoved = {name, c1, c2}

        local model = instance:getComponents()[1]:getShapeName()
        -- -2 since binary is of the same type
        nameNumbers[model] = nameNumbers[model] - 2
      else
        namesRemoved = {name}
        parentsRemoved = {name}

        local model = instance:getShapeName()
        nameNumbers[model] = nameNumbers[model] - 1
      end

      --Delete all children
      for i, celestialData in ipairs(children) do
        local inst = celestialData.instance
        local cName = inst:getName()
        local formattedName = formatObjName(cName)

        if celestialData.isParent then
          table.insert(parentsRemoved, formattedName)
        end

        table.insert(namesRemoved, formattedName)

        nameNumbers[inst:getShapeName()] = nameNumbers[inst:getShapeName()] - 1
        celestialsHandler.deleteCelestialFromInst(inst)
      end

      if instance:instanceOf() == "ClassBinarySystem" then
        celestialsHandler.deleteSystem(instance:getName())
      else
        celestialsHandler.deleteCelestial(instance:getName())
      end

      --Removes parent's orbits
      orbitGaps[instance:getName()] = nil

      --Let the UI know of all the celestials just deleted
      return {"Success", namesRemoved, parentsRemoved}
    else
      return {"Warning", "Deleting this will delete ALL orbitting celestials as well."}
    end
  else
    --Ignore binary systems
    local parentName = instance:instanceOf() ~= "ClassBinarySystem" and instance:getParentCelestial()

    --If there is a parent, get the ranges to remove from the table
    if parentName then
      local parent = celestialsHandler.findCelestial(parentName)
      if not parent then
        return {"Failure", "Could not find parent celestial with name "..(parentName or "nil")}
      end

      local radius = instance:getScaledRadius()
      local celestialSize = instance:getShapeName() == "saturn" and getActualSaturnSize(radius/0.03) or radius

      local range = getCelestialRange(parent:getPosition():getDistanceBetween(instance:getPosition()), celestialSize)
      --Remove orbital space
      adjustOrbitRangesForParent(parent, range, false)
    end

    local celestialsRemoved, parentsRemoved = {}, {}

    if instance:instanceOf() == "ClassBinarySystem" then
      local c1, c2 = unformatObjName(name)
      c1, c2 = formatObjName(c1), formatObjName(c2)
      celestialsRemoved = {name, c1, c2}
      parentsRemoved = {name, c1, c2}

      local model = instance:getComponents()[1]:getShapeName()
      -- -2 since binary is of the same type
      nameNumbers[model] = nameNumbers[model] - 2

      celestialsHandler.deleteSystem(instance:getName())
    else
      celestialsRemoved = {name}

      local model = instance:getShapeName()
      nameNumbers[model] = nameNumbers[model] - 1

      --Remove this name from the orbiting body field in the UI
      if ClassCelestial.getCelestialTypeFromModel(model) ~= "planet" then
        parentsRemoved = {name}
      end

      celestialsHandler.deleteCelestial(instance:getName(), instance:getObjectType())
    end

    return {"Success", celestialsRemoved, parentsRemoved}
  end
end

local function requestHighlightUpdate(name, binary, highlight)
  --[[
  Highlights a celestial
  ]]--
  if binary then
    local name1, name2 = unformatObjName(name)
    celestialsHandler.findCelestial(name1):highlight(highlight)
    celestialsHandler.findCelestial(name2):highlight(highlight)
  else
    name = unformatObjName(name)
    celestialsHandler.findCelestial(name):highlight(highlight)
  end
end

local function toggleAxis()
  --[[
  Shows or hides the axis visual
  ]]--
  showAxis = not showAxis
  local playerPos = be:getPlayerVehicle(0):getPosition()
  local x, y, z = playerPos.x, playerPos.y, playerPos.z

  if not scenetree.findObject("axis") then
    TorqueScript.eval([[
    new TSStatic(axis) {
      shapeName = "levels/smallgrid/art/gravitationalRacing/axis.dae";
      dynamic = "1";
      position = "]]..x..[[ ]]..y..[[ ]]..z..[[";
      scale = "0.4 0.4 0.4";
    };
    ]])
  else
    TorqueScript.eval('axis.hidden = "'..tostring(not showAxis)..'";')
    --Important to do since the player vehicle may be in a different position
    TorqueScript.eval('axis.position = "'..x..' '..y..' '..z..'";')
  end
end

local function toggleTOD()
  --[[
  Toggles between standard and midnight
  ]]--
  local STANDARD, MIDNIGHT = 0.1, 0.5
  local time = math.abs(tonumber(scenetree.findObject("tod").time) - STANDARD) <= 10^-2 and MIDNIGHT or STANDARD
  TorqueScript.eval('tod.time = "'..time..'";')
end

local function applySettings(settings)
  --[[
  Applies the settings
  ]]--
  local maxAccel = settings.acceleration
  local collision = settings.collide

  for k, instance in pairs(vehicles) do
    instance:setMaxAccel(maxAccel)
    instance:setFreeroamReset(collision == "respawn")
  end

  -- local ratio = maxAccel/200
  -- ClassCelestial.setDangerRadiiValues({51 / ratio, 101 / ratio, 151 / ratio})
  --
  -- --Update static objects
  -- celestialsHandler.callMethodOnCelestials(nil, "displayRadialZones")
end

local function initialise()
  --[[
  Creates the iniial vehicle objects
  Returns information about texture creation/detection
  ]]--
  if not initialised then
    local vehiclesInWorld = scenetree.findClassObjects("BeamNGVehicle")
    for _, vehicleName in ipairs(vehiclesInWorld) do
      vehicles[vehicleName] = ClassVehicle.new(vehicleName, scenetree.findObject(vehicleName):getPosition())
    end

    celestialsHandler.removeAllCelestials()

    --Makes it so that the checkpointsHandler knows it is not in a scenario
    checkpointsHandler.setSrcFile(nil)

    local allCelestials = ClassCelestial.getAllCelestials(false)
    for _, celestial in ipairs(allCelestials) do
      nameNumbers[celestial] = 0
    end

    -- textureHandler.createVanillaMapTextures()
    initialised = true

    local terrains = scenetree.findClassObjects("TerrainBlock")
    if #terrains > 0 then
      --This is a round-about way of getting the map name but it works
      local mapName = tostring(scenetree.findObject(terrains[1]).terrainFile):match("/(.+)/")

      if not mapName then
        return {"Failure", "Could not determine the map name!"}
      end

      if not textureHandler.hasTexturesFor(mapName) then
        textureHandler.createTextureMappings(mapName)
        --Alert the UI for the user to let them know they must restart for the textures to work
        return {"Warning", "This is a modded/unaccounted for map and so texture mappings have been created automatically by the mod. This will require a game restart to take into effect."}
      end

      return {"Success"}
    end

    if not textureHandler.hasTexturesFor(mapName) then
      return {"Failure", "No terrain exists for this map and so map name cannot be determined. (If this is a vanilla map, disregard this message.)"}
    end
  else
    return {"Success"}
  end
end

local function onVehicleSpawned(id)
  local newVehicle = be:getObjectByID(id)
  local vehicleName = newVehicle.name
  local position = newVehicle:getPosition()

  --Create vehicle instance to allow it to be affected and referenced
  vehicles[vehicleName] = ClassVehicle.new(vehicleName, newVehicle:getPosition())
end

local function onVehicleResetted(id)
  local vehicle = be:getObjectByID(id)
  local pos = vehicle:getPosition()
  --If the vehicle's spawnpoint changes, let the instance know about it
  vehicles[vehicle.name].setInitialPosition(ClassVector.new(pos.x, pos.y, pos.z))
end

local function checkDeletedVehicles()
  --[[
  Checks, and removes references to, any vehicle which has been removed
  ]]--
  for k, instance in pairs(vehicles) do
    if not instance:getObj() then
      -- if not scenetree.findObject(k) then
      vehicles[k] = nil
    end
  end
end

local function onPreRender(dt)
  --IMPORTANT: This needs to know it is in a freeroam situation everwise this is
  --called during a scenario as well, esnetially doubling the speed of update
  if initialised and not checkpointsHandler.isInScenario() then
    checkDeletedVehicles()
    celestialsHandler.update(true, vehicles, dt)
  end

  if previewData.shouldUpdate then
    local posData = previewData.posData
    local pos = posData.centerCelestial:getPosition():clone()
    pos:setY(pos:getY() + posData.orbitRadius)
    updatePreview(pos, previewData.scale, previewData.celestialType, previewData.linearPathData, previewData.binary)
  end
end

M.showPreview = showPreview
M.requestCreation = requestCreation
M.requestDeletion = requestDeletion
M.requestHighlightUpdate = requestHighlightUpdate
M.toggleAxis = toggleAxis
M.toggleTOD = toggleTOD
-- M.requestCelestialInfo = requestCelestialInfo
M.onVehicleSpawned = onVehicleSpawned
M.applySettings = applySettings
M.initialise = initialise
M.onVehicleResetted = onVehicleResetted
M.onPreRender = onPreRender
M.onScenarioChange = onScenarioChange
return M
